//   LZW  ( ).
//      .
// : 8      (512  65536 )
// : 
//
// icq:133575351  email:lawnmower-man@mail.ru

#include<stdio.h>
#include<memory.h>

// [1..256] -   
#define LZ_END	0		//   
#define LZ_START 257	//    

#define CAPABLITY 65536	//!<  

typedef  unsigned char  BYTE;
typedef  unsigned short CELL;	// 16  (   )


class LZW {
struct dict_st{
	CELL prefix;	//!< 
	CELL   next;	//!<    
	CELL  child;	//!<   
	BYTE   sym;		//!<  
};	//!<     (0        )
	int count;		//!<     
	dict_st *table;	//!<  
	void Init();					//!<  
	void Update(int,BYTE);			//!<  
	int  Find(int,BYTE);			//!<    
	void WriteSym(FILE*,int);		//!<    
	void WriteCode(FILE*,CELL);		//!<    
public:
	LZW();
	~LZW();
	void Pack(FILE*,FILE*);			//!< 
	void Unpack(FILE*,FILE*);		//!< 
};


LZW::LZW()
{
	table=new dict_st[CAPABLITY];
}

LZW::~LZW()
{
	delete[]table;
}


void LZW::Init()
{
	memset(table,0,sizeof(dict_st)*CAPABLITY);
	for(int n=0; n<256; n++) {
		table[n+1].next=n+2;
		table[n+1].sym=n;
	}
	table[n].sym=255;
	count=LZ_START;
}

//! \param code  
//! \param sym   
//! \result =true    
void LZW::Update(int code,BYTE sym)
{
	if(count>=CAPABLITY) Init();
	table[count].prefix=code;
	table[count].next=table[code].child;
	table[code].child=count;
	table[count].sym=sym;
	table[count].child=0;
	count++;
}

//! \param code  
//! \param sym     
//! \result   (0 -   )
int LZW::Find(int code,BYTE sym)
{
	code=table[code].child;
	while(code) {
		if(table[code].sym==sym) break;
		code=table[code].next;
	}
	return code;
}


//! \param in   
//! \param out  
void LZW::Pack(FILE *in,FILE *out)
{
	int code,sym,index;
	code=0;
	Init();
	for(;;) {
		sym=fgetc(in); if(sym==EOF) break;
		if(code) index=Find(code,sym); else index=sym+1;
		if(index) code=index;
		else {
			WriteCode(out,code);
			Update(code,sym);
			code=sym+1;
		}
	}
	WriteCode(out,code);
	WriteCode(out,LZ_END);
}


//! \param out   
//! \param code     
void LZW::WriteSym(FILE *out,int code)
{
	if(!code) return;
	WriteSym(out,table[code].prefix);
	fputc(table[code].sym,out);
}

//! \param out    
//! \param value  
void LZW::WriteCode(FILE *out,CELL value)
{
	fwrite(&value,1,sizeof(CELL),out);
}

//! \param in   
//! \param out  
void LZW::Unpack(FILE *in,FILE *out)
{
	CELL new_code;
	int n,old_code,sym;
	old_code=0;
	Init();
	for(;;) {
		fread(&new_code,1,sizeof(new_code),in);
		if(!new_code) break;
		if(new_code>=count) WriteSym(out,old_code),WriteSym(out,sym+1),n=old_code;
		else WriteSym(out,new_code),n=new_code;
		while(n) sym=table[n].sym,n=table[n].prefix;
		if(old_code)Update(old_code,sym);
		old_code=new_code;
	}
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	FILE *input =fopen(list[2],"rb");
	FILE *output=fopen(list[3],"wb");
	if(!input || !output) {
		printf("IO error\n\a");
		return 3;
	}
	switch(list[1][1]) {
	case 'e': LZW().Pack(input,output); break;	// 
	case 'd': LZW().Unpack(input,output); break;	// 
	}
	fclose(input);
	fclose(output);
	return 0;
}







